/**
 * @file wifi.h
 * @brief WiFi radio and interfaces definition (API header file).
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * @author anjan.chanda@iopsys.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef WIFI_H
#define WIFI_H

#include <stdint.h>
#include <stdbool.h>
#include <linux/types.h>

#ifdef __cplusplus
extern "C" {
#endif

/** API versioning stuff */
#define str(s)	_str(s)
#define _str(s)	#s
#define libwifi_version		\
	str(LIBWIFI_MAJOR)"."str(LIBWIFI_MINOR)"-"str(LIBWIFI_REV)

/*
 * APIs in this library have the following signature -
 *   int wifi_radio_xxx(char *radio_ifname, ...);
 *   int wifi_ap_xxx(char *ifname, ...);
 *   int wifi_sta_xxx(char *ifname, ...);
 *   int wifi_xxx(...);
 *
 * All the APIs return '0' on SUCCESS.
 * Any other return value means 'FAILURE'.
 *
 * In case of FAILURE, the return value is usually set to '-1'.
 */


/*
 * Sub-objects pertaining to the 'wifi' object -
 *    wifi.radio
 *    wifi.ap
 *    wifi.sta
 */

#define WIFI_NUM_AC		4
#define WIFI_NUM_ANTENNA	4
#define WIFI_NUM_WEPKEYS	4
#define WIFI_NUM_RADIUS		2

/** wifi status and reason code as per IEEE802.11 Std. */
typedef uint16_t wifi_status_t, wifi_reason_t;


/*
 * enum wifi_std - wifi standards as per IEEE802.11 Std.
 */
enum wifi_std {
	WIFI_B  = 1 << 0,        /**< 802.11b */
	WIFI_G  = 1 << 1,        /**< 802.11g */
	WIFI_A  = 1 << 2,        /**< 802.11a */
	WIFI_N  = 1 << 3,        /**< 802.11n */
	WIFI_AC = 1 << 4,        /**< 802.11ac */
	WIFI_AX = 1 << 5,        /**< 802.11ax */

	WIFI_NUM_STD = 6,
	WIFI_UNKNOWN = 1 << WIFI_NUM_STD,
};

/** enum wifi_band - frequency bands */
enum wifi_band {
	BAND_2       = 1 << 0,        /**< 0x1 for 2.4Ghz band. */
	BAND_5       = 1 << 1,        /**< 0x2 for 5Ghz band. */
	BAND_DUAL    = 3,             /**< 0x3 for both 2.4 and 5Ghz bands */
	BAND_60      = 1 << 2,        /**< 0x4 for 60Ghz */
	BAND_6       = 1 << 3,        /**< 0x8 for 6Ghz */
	BAND_UNKNOWN = 1 << 4,        /**< 0x10 (and above) for unknown band */
};

/** enum wifi_bw - bandwidth */
enum wifi_bw {
	BW20,             /**< bandwidth = 20 Mhz */
	BW40,             /**< bandwidth = 40 Mhz */
	BW80,             /**< bandwidth = 80 Mhz */
	BW160,            /**< bandwidth = 160 Mhz */
	BW_AUTO,          /**< bandwidth = ? */
	BW8080 = 7,       /**< bandwidth = 80+80 Mhz */
	BW_UNKNOWN,       /**< Unknown or Undefined */
};

/** enum wifi_chan_ext - extension channel */
enum wifi_chan_ext {
	EXTCH_NONE,     /**< no extension channel */
	EXTCH_ABOVE,    /**< extension channel above primary */
	EXTCH_BELOW,    /**< extension channel below primary */
	EXTCH_AUTO,     /**< extension channel auto */
};

/* max number of opclass per regulatory domain */
#define WIFI_NUM_OPCLASS_IN_REGDOMAIN	16

/* max number of channels per operating class */
#define WIFI_NUM_CHANNEL_IN_OPCLASS	32

/** struct wifi_opchannel - channel definition in operating class */
struct wifi_opchannel {
	int8_t txpower;               /**< max txpower in dBm */
	uint8_t num;
	uint8_t ch[WIFI_NUM_CHANNEL_IN_OPCLASS];
	//uint16_t behavior;          /**< bitflags of channel behavior */
};

/** struct wifi_opclass - operating class */
struct wifi_opclass {
	uint32_t opclass;
	uint32_t g_opclass;
	enum wifi_band band;
	enum wifi_bw bw;
	struct wifi_opchannel opchannel;
};

/** enum wifi_guard - guard interval */
enum wifi_guard {
	WIFI_GI400,
	WIFI_SGI = WIFI_GI400,         /**< 0.4us guard interval */
	WIFI_GI800,
	WIFI_LGI = WIFI_GI800,         /**< 0.8us guard interval */
	WIFI_GI1600,                   /**< 1.6us guard interval */
	WIFI_GI3200,                   /**< 3.2us guard interval */
	WIFI_GI_UNKNOWN,
	WIFI_AUTOGI = WIFI_GI_UNKNOWN, /**< guard interval auto or Unknown */
};

/** enum wifi_preamble - preamble */
enum wifi_preamble {
	SHORT_PREAMBLE,    /**< short preamble */
	LONG_PREAMBLE,     /**< long preamble */
	AUTO_PREAMBLE      /**< auto preamble */
};

/** enum wifi_bss_mode - wifi bss operational mode */
enum wifi_bss_mode {
	WIFI_ADHOC,           /**< Adhoc (or IBSS) mode of operation */
	WIFI_INFRA,           /**< Infrastructure (or AP) mode */
};

/** enum wifi_mode - wifi mode */
enum wifi_mode {
	WIFI_MODE_UNKNOWN,
	WIFI_MODE_AP,         /**< AP (or Master) mode */
	WIFI_MODE_AP_VLAN,    /**< WDS Station */
	WIFI_MODE_STA,        /**< Station (or Managed) mode */
	WIFI_MODE_MONITOR,    /**< Monitor mode */
};

/** enum wifi_security - wifi security types (user friendly names) */
enum wifi_security {
	WIFI_SECURITY_NONE,
	WIFI_SECURITY_WEP64,
	WIFI_SECURITY_WEP128,
	WIFI_SECURITY_WPAPSK,
	WIFI_SECURITY_WPA2PSK,
	WIFI_SECURITY_FT_WPA2PSK,
	WIFI_SECURITY_WPA3PSK,
	WIFI_SECURITY_WPA3PSK_T,
	WIFI_SECURITY_FT_WPA3PSK,
	WIFI_SECURITY_WPA,
	WIFI_SECURITY_WPA2,
	WIFI_SECURITY_WPA3,
};

/** @enum wifi_wpaversion - wpa versions */
enum wifi_wpaversion {
	WPA_VERSION1 = 0x100,
	WPA_VERSION2 = 0x200,
};

/** enum wifi_akm - auth and key management types */
enum wifi_akm {
	WIFI_AKM_NONE,
	WIFI_AKM_DOT1X,
	WIFI_AKM_PSK,
	WIFI_AKM_FT_DOT1X,
	WIFI_AKM_FT_PSK,           /* 4 */
	WIFI_AKM_DOT1X_SHA256,
	WIFI_AKM_PSK_SHA256,
	WIFI_AKM_TDLS,
	WIFI_AKM_SAE,              /* 8 */
	WIFI_AKM_FT_SAE,
	WIFI_AKM_AP_PEER_KEY,
	WIFI_AKM_SUITE_B,
	WIFI_AKM_SUITE_B_SHA384,   /* 12 */
	WIFI_AKM_FT_DOT1X_SHA384,
};

/** enum wifi_cipher - cipher types */
enum wifi_cipher {
	WIFI_CIPHER_GRP,
	WIFI_CIPHER_WEP40,
	WIFI_CIPHER_TKIP,
	WIFI_CIPHER_CCMP128 = 4,
	WIFI_CIPHER_CCMP = WIFI_CIPHER_CCMP128,
	WIFI_CIPHER_AES = WIFI_CIPHER_CCMP,
	WIFI_CIPHER_WEP104 = 5,
	WIFI_CIPHER_BIP_CMAC128,
	WIFI_CIPHER_NOGRP,
	WIFI_CIPHER_GCMP128 = 8,
	WIFI_CIPHER_GCMP = WIFI_CIPHER_GCMP128,
	WIFI_CIPHER_GCMP256,
	WIFI_CIPHER_CCMP256,
	WIFI_CIPHER_BIP_GMAC128,
	WIFI_CIPHER_BIP_GMAC256,
	WIFI_CIPHER_BIP_CMAC256,
};

/* authentication types bitflags */
enum {
	AUTH_OPEN    = 1<<0,
	AUTH_SHARED  = 1<<1,
	AUTH_WEPAUTO = 1<<2,
	AUTH_WPANONE = 1<<3,
	AUTH_WPA     = 1<<4,
	AUTH_WPAPSK  = 1<<5,
	AUTH_WPA2    = 1<<6,
	AUTH_WPA2PSK = 1<<7,
	AUTH_MAX = 8,
	AUTH_UNKNOWN = 1 << AUTH_MAX,
};

/* encryption types bitflags */
enum {
	CIPHER_NONE    = 1<<0,
	CIPHER_WEP     = 1<<1,
	CIPHER_TKIP    = 1<<2,
	CIPHER_AES     = 1<<3,
	CIPHER_CCMP256 = 1<<4,
	CIPHER_GCMP128 = 1<<5,
	CIPHER_GCMP256 = 1<<6,
	CIPHER_MAX = 7,
	CIPHER_UNKNOWN = 1 << CIPHER_MAX,
};

/** enum wifi_phytype - wifi phys */
enum wifi_phytype {
	PHY_DSSS   = 2,
	PHY_OFDM   = 4,
	PHY_HRDSSS = 5,
	PHY_ERP    = 6,
	PHY_HT     = 7,
	PHY_DMG    = 8,
	PHY_VHT    = 9,
};

/** @struct mimo_rate - used for phyrate calculation */
struct mimo_rate {
	uint8_t mcs;        /**< MCS value */
	uint8_t bw;         /**< Bandwidth in Mhz */
	uint8_t sgi;        /**< = 1 if SGI enabled; else 0 */
	uint8_t nss;        /**< Number of SS */
} __attribute__((packed));

/** struct wifi_rate - holds rate information */
struct wifi_rate {
	uint32_t rate;           /**< rate in Mbps */
	struct mimo_rate m;      /**< of type struct #mimo_rate */
	enum wifi_phytype phy;   /**< of type struct #wifi_phytype */
};

/*
 * The following struct wifi_caps_* defines capability bits as
 * defined in IEEE-802.11-2016 Std.
 */

/* struct wifi_caps_basic - capability field */
struct wifi_caps_basic {
	union {
		uint8_t byte[2];
		uint16_t cap;
	};
};

/* struct wifi_caps_ext - extended capabilities */
struct wifi_caps_ext {
	uint8_t byte[16];  /** upto 8 octets used now */
};

/* struct wifi_caps_ht - ht capabilities */
struct wifi_caps_ht {
	union {
		uint8_t byte[2];
		uint16_t cap;
	};
	uint8_t ampdu_params;
	uint8_t supp_mcs[16];
	union {
		uint8_t byte_ext[2];
		uint16_t cap_ext;
	};
	uint8_t txbf[4];
	uint8_t asel;
} __attribute__ ((packed));

/* struct wifi_caps_vht - vht capabilities */
struct wifi_caps_vht {
	union {
		uint8_t byte[4];
		uint32_t cap;
	};
	uint8_t supp_mcs[8];
};

/* struct wifi_caps_rm - radio measurement capability */
struct wifi_caps_rm {
	uint8_t byte[5];
};

/* struct wifi_caps_he - he capabilities */
struct wifi_caps_he {
	uint8_t byte_mac[6];
	uint8_t byte_phy[11];
	uint8_t byte_opt[46];
};

/** enum wifi_capflags - wifi capability bit-flags */
enum wifi_capflags {
	WIFI_CAP_ESS,
	WIFI_CAP_IBSS,
	WIFI_CAP_SHORT_PREAMBLE,
	WIFI_CAP_SPECTRUM_MGMT,
	WIFI_CAP_WMM,
	WIFI_CAP_SHORT_SLOT,
	WIFI_CAP_APSD,
	WIFI_CAP_RADIO_MEAS,
	WIFI_CAP_DELAYED_BA,
	WIFI_CAP_IMMEDIATE_BA,

	/* Extended capability flags */
	WIFI_CAP_2040_COEX,
	WIFI_CAP_PSMP,
	WIFI_CAP_PROXY_ARP,
	WIFI_CAP_11V_BSS_TRANS,
	WIFI_CAP_INTERWORKING,
	WIFI_CAP_TDLS,
	WIFI_CAP_QLOAD_REPORT,

	/* HT flags */
	WIFI_CAP_HT_LDPC,
	WIFI_CAP_2040,
	WIFI_CAP_HT_SMPS,
	WIFI_CAP_HT_GREENFIELD,
	WIFI_CAP_SGI20,
	WIFI_CAP_SGI40,
	WIFI_CAP_HT_TX_STBC,
	WIFI_CAP_HT_RX_STBC,
	WIFI_CAP_HT_MAX_AMSDU_3839,
	WIFI_CAP_HT_MAX_AMSDU_7935,
	WIFI_CAP_HT_CCK40,

	/* VHT related flags */
	WIFI_CAP_VHT_MPDU_3895,
	WIFI_CAP_VHT_MPDU_7991,
	WIFI_CAP_VHT_MPDU_11454,
	WIFI_CAP_160,
	WIFI_CAP_8080,
	WIFI_CAP_VHT_RX_LDPC,
	WIFI_CAP_SGI80,
	WIFI_CAP_SGI160,
	WIFI_CAP_VHT_TX_STBC,
	WIFI_CAP_VHT_RX_STBC_1SS,
	WIFI_CAP_VHT_RX_STBC_2SS,
	WIFI_CAP_VHT_RX_STBC_3SS,
	WIFI_CAP_VHT_RX_STBC_4SS,
	WIFI_CAP_VHT_SU_BFR,
	WIFI_CAP_VHT_SU_BFE,
	WIFI_CAP_VHT_MU_BFR,
	WIFI_CAP_VHT_MU_BFE,

	/* 11r FT flags */
	WIFI_CAP_FT_BSS,
	WIFI_CAP_FT_RRP,

	/* 11k RRM flags */
	WIFI_CAP_RM_LINK,
	WIFI_CAP_RM_NBR_REPORT,
	WIFI_CAP_RM_BCN_PASSIVE,
	WIFI_CAP_RM_BCN_ACTIVE,
	WIFI_CAP_RM_BCN_TABLE,
	WIFI_CAP_RM_RCPI,
	WIFI_CAP_RM_RSNI,
	WIFI_CAP_BSS_AVG_ACCESS_DELAY,
	WIFI_CAP_BSS_AVAIL_ADMISSION,

	/* HE flags */
	WIFI_CAP_HE_TWT_REQ,
	WIFI_CAP_HE_TWT_RSP,
	WIFI_CAP_HE_ALL_ACK,
	WIFI_CAP_HE_TRS,
	WIFI_CAP_HE_BSR,
	WIFI_CAP_HE_BCAST_TWT,
	WIFI_CAP_HE_32BIT_BA_BMP,
	WIFI_CAP_HE_OM_CONTROL,
	WIFI_CAP_HE_OFDMA_RA,
	WIFI_CAP_HE_AMSDU_FRAG,
	WIFI_CAP_HE_FLEX_TWT,
	WIFI_CAP_HE_QTP,
	WIFI_CAP_HE_BQR,
	WIFI_CAP_HE_SRP_RSP,
	WIFI_CAP_HE_OPS,
	WIFI_CAP_HE_AMSDU_IN_AMPDU,
	WIFI_CAP_HE_DYN_SMPS,
	WIFI_CAP_HE_40_BAND2,
	WIFI_CAP_HE_4080_BAND5,
	WIFI_CAP_HE_160_BAND5,
	WIFI_CAP_HE_160_8080_BAND5,
	WIFI_CAP_HE_242RU_BAND2,
	WIFI_CAP_HE_242RU_BAND5,
	WIFI_CAP_HE_PREAMBLE_PUNC_RX,
	WIFI_CAP_HE_LDPC_PAYLOAD,
	WIFI_CAP_HE_STBC_TX_80,
	WIFI_CAP_HE_STBC_RX_80,
	WIFI_CAP_HE_FULL_BW_UL_MUMIMO,
	WIFI_CAP_HE_PART_BW_UL_MUMIMO,
	WIFI_CAP_HE_SU_BFR,
	WIFI_CAP_HE_SU_BFE,
	WIFI_CAP_HE_MU_BFR,

	/* Custom capabilities */
	WIFI_CAP_UNASSOC_STA_MEAS,
	WIFI_CAP_UNASSOC_STA_MEAS_OFFCHANNEL,
	WIFI_CAP_STEER_LOWRSSI_STA,

	WIFI_CAP_AFTER_LAST,
	WIFI_CAP_MAX = WIFI_CAP_AFTER_LAST - 1,
};

/* Bitflags in enum wifi_capflags should be accessed through the following
 * three macros only. New bitflags may be added and/or existing ones may be
 * removed.
 *
 * Example usage -
 * a) Set WIFI_CAP_HT_TX_STBC in 'struct wifi_bss *bss' -
 *	wifi_cap_set(bss->cbitmap, WIFI_CAP_HT_TX_STBC);
 *
 * b) Print bitflags that are set in 'struct wifi_bss *bss' -
 *      wifi_for_each_cap(bss->cbitmap, flag) {
 *		if (wifi_cap_isset(bss->cbitmap, flag))
 *			printf("Capability %d is set\n", flag);
 *      }
 */

#define wifi_cap_set(b, f)	(b[(f) / 8 + 1] |= (1 << ((f) % 8)))
#define wifi_cap_isset(b, f)	(!!(b[(f) / 8 + 1] & (1 << ((f) % 8))))

#define wifi_for_each_cap(b, _iter)				\
	for (_iter = 0; _iter < WIFI_CAP_MAX; _iter++)


/** struct wifi_caps - wifi device/interface capabilities */
struct wifi_caps {
#define WIFI_CAP_BASIC_VALID	0x1
#define WIFI_CAP_EXT_VALID	0x2
#define WIFI_CAP_HT_VALID	0x4
#define WIFI_CAP_VHT_VALID	0x8
#define WIFI_CAP_RM_VALID	0x10
#define WIFI_CAP_HE_VALID	0x20
	uint32_t valid;         /** bitmap of caps available and valid */
	struct wifi_caps_basic basic;
	struct wifi_caps_ext ext;
	struct wifi_caps_ht ht;
	struct wifi_caps_vht vht;
	struct wifi_caps_rm rrm;
	struct wifi_caps_he he;
};

/** struct wifi_oper_ht - HT operation element */
struct wifi_oper_ht {
	uint8_t channel;
	uint8_t info[5];
	uint8_t basic_mcs[16];
};

/** struct wifi_oper_vht - VHT operation element */
struct wifi_oper_vht {
	uint8_t channel_width;
	uint8_t freq_mid_seg0;
	uint8_t freq_mid_seg1;
	uint8_t basic_mcs[2];
};

/** struct wifi_oper_he - HE operational element */
struct wifi_oper_he {
	uint8_t param[3];
	uint8_t color;
	uint8_t basic_mcs[2];
};

/** enum wifi_statusflags - wifi status bit-flags */
enum wifi_statusflags {
	WIFI_STATUS_WMM,                 /**< WMM enabled */
	WIFI_STATUS_PS,                  /**< in Power Save mode */
	WIFI_STATUS_LOCAL_OPCLASS,       /**< in local regulatory opclass */
	WIFI_STATUS_COLOR,               /**< BSS color enabled */

	WIFI_STATUS_AFTER_LAST,
	WIFI_STATUS_MAX = WIFI_STATUS_AFTER_LAST - 1,
};

#define wifi_status_set(b, f)	(b[(f) / 8 + 1] |= (1 << ((f) % 8)))
#define wifi_status_isset(b, f)	(!!(b[(f) / 8 + 1] & (1 << ((f) % 8))))

#define wifi_for_each_status(b, _iter)				\
	for (_iter = 0; _iter < WIFI_STATUS_MAX; _iter++)


/** struct wifi_iface - interface per wifi radio */
struct wifi_iface {
	char name[16];
	enum wifi_mode mode;
};

/* limit max interfaces per radio */
#define WIFI_IFACE_MAX_NUM	16


/*
 * struct wifi_radio_stats - per radio statistics
 */
struct wifi_radio_stats {
	unsigned long tx_bytes;
	unsigned long rx_bytes;
	unsigned long tx_pkts;
	unsigned long rx_pkts;
	uint32_t tx_err_pkts;
	uint32_t rx_err_pkts;
	uint32_t tx_dropped_pkts;
	uint32_t rx_dropped_pkts;
	uint32_t rx_plcp_err_pkts;
	uint32_t rx_fcs_err_pkts;
	uint32_t rx_mac_err_pkts;
	uint32_t rx_unknown_pkts;
	int noise;
};

/** struct wifi_radio_diagnostic - radio diagnostic data in latest second */
struct wifi_radio_diagnostic {
	uint32_t channel_busy;     /** in usecs */
	uint32_t tx_airtime;       /** in usecs */
	uint32_t rx_airtime;       /** in usecs */
	uint32_t obss_airtime;     /** in usecs */
	uint32_t cca_time;         /** in usecs */
	uint32_t false_cca_count;  /** number of false detections (glitches) */
};

/*
 * struct wifi_radio - wifi radio struct
 */
struct wifi_radio {
	bool enabled;
	uint8_t tx_streams;
	uint8_t rx_streams;
	uint32_t max_bitrate;
	enum wifi_band oper_band;	/**< exactly one band from supp_band */
	uint8_t supp_band;		/**< bitmap of wifi frequency bands */
	uint8_t supp_std;		/**< bitmap of wifi_std */
	uint8_t oper_std;		/**< bitmap of wifi_std from supp_std */
	uint8_t channel;		/**< current primary (ctrl) channel */
	uint8_t supp_channels[64];
	uint8_t oper_channels[64];	/**< in use channels */
	bool acs_capable;
	bool acs_enabled;
	uint32_t acs_interval;		/**< in secs */
	uint32_t supp_bw;               /**< bitmap of wifi_bw */
	enum wifi_bw curr_bw;
	enum wifi_chan_ext extch;       /**< current extension channel */
	enum wifi_guard gi;
	int8_t txpower;                 /**< -1 for auto; else in %-age */
	int8_t txpower_dbm;             /**< in dBm */
	bool dot11h_capable;
	bool dot11h_enabled;
	char regdomain[4];              /**< countrycode + "O" | "I" + NUL */
	uint8_t srl;                    /**< short retry limit */
	uint8_t lrl;                    /**< long retry limit */
	uint32_t frag;
	uint32_t rts;
	uint32_t beacon_int;            /**< in msecs */
	uint32_t dtim_period;
	bool aggr_enable;
	enum wifi_preamble pr;
	uint32_t basic_rates[32];
	uint32_t oper_rates[32];
	uint32_t supp_rates[32];
	struct wifi_radio_stats stats;
	struct wifi_radio_diagnostic diag;
	uint8_t max_iface_ap;
	uint8_t max_iface_sta;
	uint8_t num_iface;
	struct wifi_iface iface[WIFI_IFACE_MAX_NUM];
};

/*
 * struct wifi_ap_stats - per-BSS statistics
 */
struct wifi_ap_stats {
	unsigned long tx_bytes;
	unsigned long rx_bytes;
	unsigned long tx_pkts;
	unsigned long rx_pkts;
	uint32_t tx_err_pkts;
	uint32_t tx_rtx_pkts;
	uint32_t tx_rtx_fail_pkts;
	uint32_t tx_retry_pkts;
	uint32_t tx_mretry_pkts;
	uint32_t ack_fail_pkts;
	uint32_t aggr_pkts;
	uint32_t rx_err_pkts;
	unsigned long tx_ucast_pkts;
	unsigned long rx_ucast_pkts;
	uint32_t tx_dropped_pkts;
	uint32_t rx_dropped_pkts;
	unsigned long tx_mcast_pkts;
	unsigned long rx_mcast_pkts;
	unsigned long tx_bcast_pkts;
	unsigned long rx_bcast_pkts;
	unsigned long rx_unknown_pkts;
};

/*
 * enum wifi_ap_confstatus - AP's configuration status
 */
enum wifi_ap_confstatus {
	WIFI_AP_DISABLED,
	WIFI_AP_ENABLED,
	WIFI_AP_ERR_MISCONF,
	WIFI_AP_ERROR,
};

/*
 * enum wifi_mfp_config - management frame protection config
 */
enum wifi_mfp_config {
	WIFI_MFP_DISABLED,
	WIFI_MFP_OPTIONAL,
	WIFI_MFP_REQUIRED,
};

struct wifi_rsne {
	uint16_t wpa_versions;                /**< bitmap of WPA_VERSION* */
	uint32_t group_cipher;                /**< one of WIFI_CIPHER_* */
	uint32_t pair_ciphers;                /**< bitmap of WIFI_CIPHER_* */
	uint32_t akms;                        /**< bitmap of WIFI_AKM_* */
	uint16_t rsn_caps;
};

struct wifi_ap_security {
	uint32_t supp_modes;    /** bitmap of supported WIFI_SECURITY_* */
	uint32_t curr_mode;     /** from wifi_rsnie in beacon/probe-resp */
	uint8_t wepidx;
	uint8_t wep104[WIFI_NUM_WEPKEYS][13];
	uint8_t wep40[WIFI_NUM_WEPKEYS][5];
	uint8_t psk[32];
	char passphrase[64];
	uint32_t rekey_int;
	ipaddress_t radius_server[WIFI_NUM_RADIUS];
	uint32_t radius_port[WIFI_NUM_RADIUS];
	char radius_secret[WIFI_NUM_RADIUS][128];
	enum wifi_mfp_config mfp;
};

/* wps configuration state */
enum wps_state {
	WPS_DISABLED,
	WPS_UNCONFIGURED,
	WPS_CONFIGURED,
	WPS_SETUPLOCKED,
};

/*
 * enum wps_method - WPS configuration methods
 */
enum wps_method {
	WPS_METHOD_PBC,
	WPS_METHOD_PIN,
	WPS_METHOD_UNKNOWN,
};

#if 1	// redefine
/* wpa status */
enum wps_status {
	WPS_STATUS_IDLE,
	WPS_STATUS_INIT = WPS_STATUS_IDLE,
	WPS_STATUS_PROCESSING,
	WPS_STATUS_SUCCESS,
	WPS_STATUS_FAIL,
	WPS_STATUS_TIMEOUT,
	WPS_STATUS_UNKNOWN,
};
#endif

struct wifi_ap_wps {
	bool enable;
	uint32_t supp_methods;     /** bitmap of enum wps_method */
	enum wps_method en_method;
	enum wps_status status;
	uint32_t version;
	char pin[8];    /** 4 or 8 digit number only; not NUL-terminated */
};

/*
 * struct wps_device - wps device info
 */
struct wps_device {
	char name[32];
	char manufacturer[64];
	char modelname[32];
	char modelnum[32];
	char serialnum[32];
};

/*
 * struct wifi_sta_stats - per STA statistics
 */
struct wifi_sta_stats {
	uint64_t tx_bytes;
	uint64_t rx_bytes;
	uint32_t tx_pkts;
	uint32_t rx_pkts;
	uint32_t tx_err_pkts;
	uint32_t tx_rtx_pkts;
	uint32_t tx_rtx_fail_pkts;
	uint32_t tx_retry_pkts;
	uint32_t tx_mretry_pkts;
	uint32_t tx_fail_pkts;
	uint64_t rx_fail_pkts;
};

/*
 * struct wifi_sta - info of a wifi sta from AP's assoclist
 */
struct wifi_sta {
	uint8_t macaddr[6];         /**< sta macaddress */
	uint8_t bssid[6];           /**< bssid of connected ap */
	uint8_t sbitmap[4];         /**< bitmap of enum wifi_statusflags */
	uint8_t cbitmap[16];        /**< bitmap for enum wifi_capflags */
	struct wifi_caps caps;      /**< capabilities */
	uint8_t oper_std;           /**< bitmap of wifi_std from supp_std */
	uint32_t maxrate;           /**< max phy operational rate in Mbps */
	struct wifi_rate rx_rate;   /**< latest rate: from AP -> this STA */
	struct wifi_rate tx_rate;   /**< latest rate: this STA -> AP */
	uint32_t rx_thput;          /**< AP -> this STA instant throughput */
	uint32_t tx_thput;          /**< this STA -> AP instant throughput */
	int8_t rssi_avg;                /**< average rssi */
	int8_t rssi[WIFI_NUM_ANTENNA];  /**< latest rssi in dBm per-chain */
	int8_t noise_avg;               /**< average phy noise in dBm */
	int8_t noise[WIFI_NUM_ANTENNA]; /**< latest noise in dBm */
	struct wifi_sta_stats stats;
	uint64_t tx_airtime;        /**< Tx airtime(msecs) in the last second */
	uint64_t rx_airtime;        /**< Rx airtime(msecs) in the last second */
	int8_t airtime;             /**< airtime in %-age in the last second */
	uint32_t conn_time;         /**< time in secs since connected */
	uint32_t idle_time;         /**< inactive time in secs */
};

/*
 * struct wifi_monsta_config - monitored sta config
 */
struct wifi_monsta_config {
	bool enable;			/**< enable/disable STA monitor */
};

/*
 * struct wifi_monsta - monitored sta info
 */
struct wifi_monsta {
	uint8_t macaddr[6];
	int8_t rssi[WIFI_NUM_ANTENNA];   /**< latest rssi in dBm */
	int8_t rssi_avg;		 /** < average rssi */
	int last_seen;			 /** < last seen in seconds */
	struct wifi_caps caps;		 /**< capabilities */
};

/*
 * struct wifi_ap_wmm_ac_stats - AP's per WMM_AC statistics
 */
struct wifi_ap_wmm_ac_stats {
	uint64_t tx_bytes;
	uint64_t rx_bytes;
	uint32_t tx_pkts;
	uint32_t rx_pkts;
	uint32_t tx_err_pkts;
	uint32_t rx_err_pkts;
	uint32_t tx_rtx_pkts;
};

/** enum wmm_ac_type - wmm access categories */
enum wmm_ac_type {
	BK,
	BE,
	VI,
	VO,
};

/*
 * struct wifi_ap_wmm_ac - AP's per WMM_AC configuration
 */
struct wifi_ap_wmm_ac {
	enum wmm_ac_type ac;
	uint8_t aifsn;
	uint8_t cwmin;
	uint8_t cwmax;
	uint8_t txop;
	bool ack_policy;
	struct wifi_ap_wmm_ac_stats stats;
};

/*
 * struct wifi_ap_accounting - accounting server info
 */
struct wifi_ap_accounting {
	bool enable;
	ipaddress_t server[WIFI_NUM_RADIUS];
	uint32_t server_port[WIFI_NUM_RADIUS];
	char secret[WIFI_NUM_RADIUS][128];
	uint32_t intm_interval;
};

enum acl_policy {
	ALLOW,
	DENY,
};

struct wifi_ap_acl {
	bool acl_enabled;
	enum acl_policy policy;
	void *allowlist;           /** points to array of STA macaddress */
	void *denylist;            /** points to array of STA macaddress */
};

/** struct wifi_ap_load - Bss load */
struct wifi_ap_load {
	uint16_t sta_count;         /**< number of STAs connected */
	uint8_t utilization;        /**< channel utilization [0..255] */
	uint16_t available;         /**< available admission capacity */
} __attribute__((packed));

/*
 * struct wifi_bss - BSS information.
 * This struct is also used to store info of a scanned AP.
 */
struct wifi_bss {
	uint8_t ssid[33];
	uint8_t bssid[6];
	enum wifi_bss_mode mode;
	uint8_t channel;
	enum wifi_bw curr_bw;
	enum wifi_band band;
	uint8_t supp_std;
	uint8_t oper_std;
	int rssi;
	int noise;
	struct wifi_rsne rsn;
	uint32_t auth;		/* replacing 'sec'; will be deprecated */
	uint32_t enc;		/* deprecated; 'g_enc' is now 'rsn.group_cipher' */
	uint32_t security;      /**< bitmap of enum wifi_security */
	uint32_t beacon_int;
	uint32_t dtim_period;
	struct wifi_ap_load load;
	struct wifi_caps caps;
	uint8_t cbitmap[16];        /**< bitmap for enum wifi_capflags */
} __attribute__ ((packed));

/*
 * struct wifi_bss_detail - details of scanned neighbor BSS
 */
struct wifi_bss_detail {
	struct wifi_bss basic;
	uint32_t ielen;
	uint8_t ie[1024];
};

/*
 * struct wifi_ap - AP structure corresponding to a BSS
 */
struct wifi_ap {
	struct wifi_bss bss;
	enum wifi_ap_confstatus confstatus;
	ifopstatus_t opstatus;
	bool ssid_advertised;
	bool wmm_cap;
	bool uapsd_cap;
	bool wmm_enabled;
	bool uapsd_enabled;
	uint32_t assoclist_max;
	bool isolate_enabled;
	struct wifi_ap_acl acl;
	struct wifi_ap_security sec;
	struct wifi_ap_wps wps;
	struct wifi_ap_accounting acct;
	struct wifi_ap_wmm_ac ac[WIFI_NUM_AC];
	struct wifi_ap_stats stats;
	uint32_t assoclist_num;
	void *assoclist;  /** list of struct wifi_sta or simply macaddresses */
};

enum wifi_scan_state {
	SCAN_NONE,
	SCAN_REQUESTED,
	SCAN_SCANNING = SCAN_REQUESTED,
	SCAN_CANCELED,
	SCAN_COMPLETE,
	SCAN_DONE = SCAN_COMPLETE,
	SCAN_ERROR,
};

/*
 * struct wifi_neighbor - struct holding scan results of neighbor APs
 */
struct wifi_neighbor {
	char radio[16];               /** scanning wifi radio device name */
	uint32_t num_result;          /** num of scanned APs */
	struct wifi_bss *scanreslist; /** points to array of struct wifi_bss */
};

/*
 * struct wifi - structure for a 'wifi' object
 */
struct wifi {
	uint32_t num_radio;
	uint32_t num_ap;
	struct wifi_radio *radiolist; /** points to struct wifi_radio array */
	struct wifi_ap *aplist;       /** points to struct wifi_ap array */
};

/** struct scan_param - scan request parameters */
struct scan_param {
	char ssid[33];     /**< ssid specific scan */
	uint8_t bssid[6];  /**< scan bssid */
	uint32_t channel;  /**< channel to scan */
	uint8_t type;      /**< auto ( = 0), active (= 1), passive (=2) */
};

/** struct acs_param - auto channel sel arguments */
struct acs_param {
	// TODO
};

/** struct chan_switch_param - channel switch parameters */
struct chan_switch_param {
	int count;		/**< number of beacons before switch */
	int freq;		/**< control channel frequency */
	int bandwidth;		/**< bandwidth in MHz */
	int sec_chan_offset;	/**< for HT40+/HT40- */
	int cf1;		/**< central frequency1 */
	int cf2;		/**< central frequency2 */
	bool blocktx;		/**< blocktx during channel switch */
	bool ht;		/**< use HT */
	bool vht;		/**< use VHT */
};

#define OUI_LEN		3

/** struct vendor_ie - vendor ie struct */
struct vendor_ie {
	struct {
		__u8 eid;
		__u8 len;
	} ie_hdr;
	__u8 oui[OUI_LEN];
	__u8 data[];
} __attribute__ ((packed));

/** struct vendor_iereq - vendor specific ie request struct */
struct vendor_iereq {
	__u32 mgmt_subtype;  /**< bitmap of management frame subtypes */
	struct vendor_ie ie; /**< vendor ie structure */
};

/* bitmap of wps roles */
enum wps_role {
	WPS_ENROLLEE  = 1<<0,
	WPS_PROXY     = 1<<1,
	WPS_REGISTRAR = 1<<2,
	WPS_ROLE_MAX  = 1<<3,
};

/**
 * struct wps_param - WPS parameter to be used during registration
 * @role: enrollee, registrar or proxy. Bitmap of %WPS_ENROLLEE,
 *	%WPS_REGISTRAR, %WPS_PROXY etc.
 * @method: WPS configuration method, i.e. one of enum wps_method
 * @pin: pin value if wps_method 'PIN' is used for registration
 */
struct wps_param {
	enum wps_role role;		/**< bitmap of wps_role */
	enum wps_method method;		/**< bitmap of wps_method */
	union {
		unsigned long pin;	/**< pin value for PIN method */
	};
};

/* struct vlan_param */
struct vlan_param {
	uint8_t dir;	/* 0 = ingress, 1 = egress */
	uint8_t pcp;	/* priority */
	uint16_t vid;	/* vlan id */
};

/* 11r related structs and defines */

#define FT_R0KH_ID_MAX_LEN	48
#define FT_R1KH_ID_LEN		6
#define WPA_PMK_NAME_LEN	16
#define PMK_LEN			32

struct fbt_keys {
	uint8_t ap_address[6];                /** bssid */
	uint8_t r1kh_id[FT_R1KH_ID_LEN];
	uint8_t s1kh_id[6];                   /** mac address of sta */
	uint8_t pmk_r0_name[WPA_PMK_NAME_LEN];
	uint8_t pmk_r1[PMK_LEN];
	uint8_t pmk_r1_name[WPA_PMK_NAME_LEN];
	uint8_t r0kh_id[FT_R0KH_ID_MAX_LEN];
	uint8_t r0kh_id_len;
	uint16_t pairwise;
} __attribute__((packed));


/* 11k neighbor BSS related structs and defines */

/** struct nbr_header - meta data for 'struct nbr' */
struct nbr_header {
#define NBR_F_EXCLUDE	0x1  /** exclude this neighbor */
#define NBR_F_USER	0x2  /** user added entry */
	uint32_t flags;
} __attribute__((packed));

/** enum nbr_bssinfo - 802.11k neighbor bssid info bitflags */
enum nbr_bssinfo {
	NBR_BSSINFO_REACHABLE = 1<<0,
	NBR_BSSINFO_SECURITY  = 1<<1,
	NBR_BSSINFO_QOS       = 1<<2,
	NBR_BSSINFO_RRM       = 1<<3,
	NBR_BSSINFO_HT        = 1<<4,
	NBR_BSSINFO_VHT       = 1<<5,
};

/*
 * struct nbr - 802.11k neighbor BSS
 */
struct nbr {
	//struct nbr_header h; /**< header containing meta data */
	uint8_t bssid[6];    /**< Bssid */
	uint32_t bssid_info; /**< as in IEEE 802.11-2016 9.4.2.37 */
	uint8_t reg;         /**< regulatory region */
	uint8_t channel;     /**< channel */
	uint8_t phy;         /**< of enum wifi_phytype */
	/* ... optional subelms */
} __attribute__((packed));

/*
 * struct sta_nbr - 802.11k neighbor BSS based on beacon report from STA
 */
struct sta_nbr {
	uint8_t bssid[6];
	int8_t rssi;
	int8_t rsni;
} __attribute__((packed));

/*
 * struct rrm_measurement_beacon_request - 11k beacon request
 */
struct rrm_measurement_beacon_request {
	uint8_t oper_class;		/**< Operating Class */
	uint8_t channel;		/**< Channel Number */
	uint16_t rand_interval;		/**< Randomization Interval (in TUs) */
	uint16_t duration;		/**< Measurement Duration (in TUs) */
	uint8_t mode;			/**< Measurement Mode */
	uint8_t bssid[6];		/**< BSSID */
	uint8_t variable[0];		/**< Optional Subelements */
} __attribute__((packed));

/** BTM Request mode */
enum {
	WIFI_BTMREQ_PREF_INC     = 1 << 0,
	WIFI_BTMREQ_ABRIDGED     = 1 << 1,
	WIFI_BTMREQ_DISASSOC_IMM = 1 << 2,
	WIFI_BTMREQ_BSSTERM_INC  = 1 << 3,
	WIFI_BTMREQ_ESSTERM_IMM  = 1 << 4,
};

/*
 * struct wifi_btmreq - 11v BTM Request
 */
struct wifi_btmreq {
	uint8_t mode;		/**< bitmap of WIFI_BTMREQ_* */
	uint16_t disassoc_tmo;	/**< in tbtts when DISASSOC_IMM is set */
	uint8_t validity_int;	/**< in tbtts until candidate list is valid */
	uint16_t bssterm_dur;	/**< bss termination duration in minutes */
};

#define WIFI_FRAME_MGMT               0
#define WIFI_FRAME_CTRL               1
#define WIFI_FRAME_DATA               2

/* management */
#define WIFI_FRAME_ASSOC_REQ         0
#define WIFI_FRAME_ASSOC_RESP        1
#define WIFI_FRAME_REASSOC_REQ       2
#define WIFI_FRAME_REASSOC_RESP      3
#define WIFI_FRAME_PROBE_REQ         4
#define WIFI_FRAME_PROBE_RESP        5
#define WIFI_FRAME_BEACON            8
#define WIFI_FRAME_DISASSOC          10
#define WIFI_FRAME_AUTH              11
#define WIFI_FRAME_DEAUTH            12
#define WIFI_FRAME_ACTION            13

struct radio_entry {
	char name[16];
};

struct iface_entry {
	char name[16];
	enum wifi_mode mode;
};

/**
 * @struct wifi_radio_ops
 * @brief wifi radio related operations.
 *
 * All radio/device specific operations are handled through this structure.
 * In order to support a new wifi chipset, struct wifi_radio_ops alongwith
 * struct wifi_iface_ops must be implemented by its driver module.
 *
 * Unless otherwise mentioned, the following functions return 0 on Success,
 * and -1 on Failure.
 *
 * <b>int (*list)(struct radio_entry *radio, int *num).</b>\n
 *	@brief            Get list of preset radios
 *	@param[out] radio  radio array
 *	@param[out] num   number of entries in radio array
 *
 * <b>int (*info)(const char *name, struct wifi_radio *radio).</b>\n
 *	@brief            Get information about the radio interface.
 *	@param[in] name   radio interface name
 *	@param[out] radio struct wifi_radio
 *
 * <b>int (*get_supp_band)(const char *name, uint32_t *bands)</b>\n
 *	@brief            Get supported WiFi bands in bands param.
 *	@param[in] name   radio interface name
 *	@param[out] bands bitmap of bands from struct wifi_band
 *
 * <b>int (*get_oper_band)(const char *name, enum wifi_band *band)</b>\n
 *	@brief            Get current operating WiFi band.
 *	@param[in] name   radio interface name
 *	@param[out] band  band struct wifi_band type
 *
 * <b>int (*get_caps)(const char *name, struct wifi_caps *caps)</b>\n
 *	@brief            Get WiFi radio capabilities.
 *	@param[in] name   radio interface name
 *	@param[out] caps  capabilities struct wifi_caps
 *
 * <b>int (*get_supp_stds)(const char *name, uint8_t *std)</b>\n
 *	@brief            Get WiFi supported standards.
 *	@param[in] name   radio interface name
 *	@param[out] std   bitmap of #enum wifi_std
 *
 * <b>int (*get_oper_stds)(const char *name, uint8_t *std)</b>\n
 *	@brief            Get WiFi operational standards.
 *	@param[in] name   radio interface name
 *	@param[out] std   bitmap of enum wifi_std
 *
 * <b>int (*get_country)(const char *name, char *alpha2)</b>\n
 *	@brief             Get operating country information.
 *	@param[in] name    radio interface name
 *	@param[out] alpha2 country code
 *
 * <b>int (*get_channel)(const char *ifname, uint32_t *channel, enum wifi_bw *bw)</b>\n
 *	@brief              Get operating channel information.
 *	@param[in] ifname   radio interface name
 *	@param[out] channel primary control channel
 *	@param[out] bw      channel bandwidth
 *
 * <b>int (*set_channel)(const char *ifname, uint32_t channel, enum wifi_bw bw)</b>\n
 *	@brief              Set operating channel with bandwidth.
 *	@param[in] ifname   radio interface name
 *	@param[out] channel primary control channel
 *	@param[out] bw      channel bandwidth
 *
 * <b>int (*get_supp_channels)(const char *name, uint32_t *chlist, int *num,
 *		const char *alpha2, enum wifi_band f, enum wifi_bw bw)</b>\n
 *	@brief              Get supported channels.
 *	@param[in] name     radio interface name
 *	@param[out] chlist  array of channels
 *	@param[out] num     number of channels in chlist array
 *	@param[in] alpha2   country code
 *	@param[in] f        frequency band #enum wifi_band
 *	@param[in] bw       channel bandwidth enum wifi_bw
 *
 * <b>int (*get_oper_channels)(const char *name, uint32_t *chlist, int *num,
 *		const char *alpha2, enum wifi_band f, enum wifi_bw b)</b>\n
 *	@brief              Get operating channels.
 *	@param[in] name     radio interface name
 *	@param[out] chlist  array of channels
 *	@param[out] num     number of channels in chlist array
 *	@param[in] alpha2   country code
 *	@param[in] f        frequency band #enum wifi_band
 *	@param[in] bw       channel bandwidth enum wifi_bw
 *
 * <b>int (*get_supp_opclass)(const char *name, int *num,
 *						struct wifi_opclass *o)</b>\n
 *	@brief              Get supported operating classes.
 *	@param[in] name     radio interface name
 *	@param[out] num     number of operating classes supported
 *	@param[out] o       array of struct wifi_opclass elements
 *
 * <b>int (*get_curr_opclass)(const char *name, int *num,
 *						struct wifi_opclass *o)</b>\n
 *	@brief              Get current operating class(es).
 *	@param[in] name     radio interface name
 *	@param[out] num     number of current operating classes
 *	@param[out] o       array of struct wifi_opclass elements
 *
 * <b>int (*get_bandwidth)(const char *name, enum wifi_bw *bw)</b>\n
 *	@brief             Get operating channel bandwidth.
 *	@param[in] name    radio interface name
 *	@param[out] bw     bandwidth #enum wifi_bw
 *
 * <b>int (*get_maxrate)(const char *name, unsigned long *rate)</b>\n
 *	@brief             Get maximum supported phy rate.
 *	@param[in] name    radio interface name
 *	@param[out] rate   rate in Mbps
 *
 * <b>int (*get_basic_rates)(const char *name, int *num, uint32_t *rates)</b>\n
 *	@brief             Get basic phy rates.
 *	@param[in] name    radio interface name
 *	@param[out] num    number of elements in rates array
 *	@param[out] rates  array of rates in Mbps
 *
 * <b>int (*get_oper_rates)(const char *name, int *num, uint32_t *rates)</b>\n
 *	@brief             Get operational phy rates.
 *	@param[in] name    radio interface name
 *	@param[out] num    number of elements in rates array
 *	@param[out] rates  array of rates in Mbps
 *
 * <b>int (*get_supp_rates)(const char *name, int *num, uint32_t *rates)</b>\n
 *	@brief             Get supported phy rates.
 *	@param[in] name    radio interface name
 *	@param[out] num    number of elements in rates array
 *	@param[out] rates  array of rates in Mbps
 *
 * <b>int (*get_stats)(const char *ifname, struct wifi_radio_stats *s)</b>\n
 *	@brief             Get radio statistics.
 *	@param[in] ifname  radio interface name
 *	@param[out] s      radio stats and counters
 *
 * <b>int (*scan)(const char *name, struct scan_param *p)</b>\n
 *	@brief             Trigger scanning.
 *	@param[in] name    radio interface name
 *	@param[in] p       scan request parameters
 *
 * <b>int (*get_scan_results)(const char *name, struct wifi_bss *bsss, int *num)</b>\n
 *	@brief             Get scan results.
 *	@param[in] name    radio interface name
 *	@param[out] bsss   array of scanned APs
 *	@param[out] num    number of scanned APs
 *
 * <b>int (*get_bss_scan_result)(const char *name, uint8_t *bssid,
 *				struct wifi_bss_detail *b)</b>\n
 *	@brief             Get scan result details of a specific AP.
 *	@param[in] name    radio interface name
 *	@param[in] bssid   bssid of a scanned AP
 *	@param[out] b      scan result including IE details
 *
 * <b>int (*get_noise)(const char *ifname, int *noise);
 *	@brief             Get current noise value.
 *	@param[in] name    radio interface name
 *	@param[out] noise  noise value in dBm
 *
 * <b>int (*acs)(const char *name, struct acs_param *p)</b>\n
 *	@brief             Trigger ACS (auto channel selection).
 *	@param[in] name    radio interface name
 *	@param[in] p       ACS request parameters
 *
 * <b>int (*get_param)(const char *name, const char *param, int *len, void *val)</b>\n
 *	@brief             Get radio parameter value(s).
 *	@param[in] name    radio interface name
 *	@param[in] param   radio parameter name
 *	@param[out] len    length of the returned parameter value
 *	@param[out] val    parameter value
 *
 * <b>int (*set_param)(const char *name, const char *param, int len, void *val)</b>\n
 *	@brief             Set radio parameter value(s).
 *	@param[in] name    radio interface name
 *	@param[in] param   radio parameter name
 *	@param[in] len     length of the parameter
 *	@param[in] val     value of parameter
 *
 * <b>int (*add_iface)(const char *name, enum wifi_mode m, char *argv[])</b>\n
 *	@brief             Create a WiFi interface on this radio.
 *	@param[in] name    radio interface name
 *	@param[in] m       wifi mode f.e. WIFI_MODE_AP, WIFI_MODE_STA etc.
 *	@param[in] argv    string arguments array of wifi attributes and values
 *
 * <b>int (*del_iface)(const char *name, const char *ifname)</b>\n
 *	@brief             Delete a WiFi interface on this radio.
 *	@param[in] name    radio interface name
 *	@param[in] ifname  wifi interface name to be deleted
 *
 * <b>int (*list_iface)(const char *name, struct iface_entry *iface, int *num)</b>\n
 *	@brief             Delete a WiFi interface on this radio.
 *	@param[in] name    radio interface name
 *	@param[out] iface  array of interfaces
 *	@param[out] num    number of entries in iface array
 */
struct wifi_radio_ops {
	int (*info)(const char *name, struct wifi_radio *radio);
	int (*get_supp_band)(const char *name, uint32_t *bands);
	int (*get_oper_band)(const char *name, enum wifi_band *band);
	int (*get_caps)(const char *name, struct wifi_caps *caps);
	int (*get_supp_stds)(const char *name, uint8_t *std);
	int (*get_oper_stds)(const char *name, uint8_t *std);

	int (*get_country)(const char *name, char *alpha2);
	int (*get_channel)(const char *ifname, uint32_t *channel, enum wifi_bw *bw);
	int (*set_channel)(const char *ifname, uint32_t channel, enum wifi_bw bw);
	int (*get_supp_channels)(const char *name, uint32_t *chlist, int *num,
			const char *alpha2, enum wifi_band f, enum wifi_bw b);
	int (*get_oper_channels)(const char *name, uint32_t *chlist, int *num,
			const char *alpha2, enum wifi_band f, enum wifi_bw b);

	int (*get_supp_opclass)(const char *name, int *num_opclass,
						struct wifi_opclass *o);

	int (*get_curr_opclass)(const char *name, struct wifi_opclass *o);

	int (*get_bandwidth)(const char *name, enum wifi_bw *bw);
	int (*get_maxrate)(const char *name, unsigned long *rate_Mbps);
	int (*get_basic_rates)(const char *name, int *num, uint32_t *rates_kbps);
	int (*get_oper_rates)(const char *name, int *num, uint32_t *rates_kbps);
	int (*get_supp_rates)(const char *name, int *num, uint32_t *rates);
	int (*get_stats)(const char *ifname, struct wifi_radio_stats *s);

	int (*scan)(const char *name, struct scan_param *p);
	int (*get_scan_results)(const char *name, struct wifi_bss *bsss, int *num);
	int (*get_bss_scan_result)(const char *name, uint8_t *bssid,
						struct wifi_bss_detail *b);

	int (*get_noise)(const char *ifname, int *noise);

	int (*acs)(const char *name, struct acs_param *p);

	int (*get_param)(const char *name, const char *param, int *len, void *val);
	int (*set_param)(const char *name, const char *param, int len, void *val);

	int (*add_iface)(const char *name, enum wifi_mode m, char *argv[]);
	int (*del_iface)(const char *name, const char *ifname);
	int (*list_iface)(const char *name, struct iface_entry *iface, int *num);
};


/**
 * @struct wifi_iface_ops
 * @brief WiFi interface related operations.
 *
 * BSS/STA interface operations are handled through this structure.
 *
 * <b>int (*start_wps)(const char *ifname, struct wps_param wps)</b>\n
 *	@brief             Start WPS registration
 *	@param[in] ifname  interface name
 *	@param[in] wps     wps_param structure
 *
 * <b>int (*stop_wps)(const char *ifname)</b>\n
 *	@brief             Stop ongoing WPS registration
 *	@param[in] ifname  interface name
 *
 * <b>int (*get_wps_status)(const char *ifname, enum wps_status *s)</b>\n
 *	@brief             Get latest wps registration status
 *	@param[in] ifname  interface name
 *	@param[out] s      wps_param structure
 *
 * <b>int (*get_wps_pin)(const char *ifname, unsigned long *pin)</b>\n
 *	@brief             Get AP's (i.e. own) WPS pin
 *	@param[in] ifname  interface name
 *	@param[out] pin    wps pin value
 *
 * <b>int (*set_wps_pin)(const char *ifname, unsigned long pin)</b>\n
 *	@brief             Set AP's (i.e. own) WPS pin
 *	@param[in] ifname  interface name
 *	@param[in] pin     wps pin value
 *
 * <b>int (*get_wps_device_info)(const char *ifname, struct wps_device *s)</b>\n
 *	@brief             Get WPS device information
 *	@param[in] ifname  interface name
 *	@param[out] s      wps_device structure
 *
 * <b>int (*get_caps)(const char *ifname, struct wifi_caps *caps)</b>\n
 *	@brief             Get capabilities
 *	@param[in] ifname  interface name
 *	@param[out] caps   wifi_caps structure
 *
 * <b>int (*get_mode)(const char *ifname, enum wifi_mode *mode)</b>\n
 *	@brief             Get WiFi mode
 *	@param[in] ifname  interface name
 *	@param[out] mode   WiFi mode f.e. WIFI_MODE_AP or WIFI_MODE_STA etc.
 *
 * <b>int (*get_security)(const char *ifname, uint32_t *auth, uint32_t *enc)</b>\n
 *	@brief             Get security info
 *	@param[in] ifname  interface name
 *	@param[out] auth   authtication type
 *	@param[out] enc    encryption type
 *
 * <b>int (*add_vendor_ie)(const char *ifname, struct vendor_iereq *req)</b>\n
 *	@brief             Add vendor specific ie element
 *	@param[in] ifname  interface name
 *	@param[in] req     vendor_iereq structure
 *
 * <b>int (*del_vendor_ie)(const char *ifname, struct vendor_iereq *req)</b>\n
 *	@brief             Delete vendor specific ie element
 *	@param[in] ifname  interface name
 *	@param[in] req     vendor_iereq structure
 *
 * <b>int (*get_vendor_ies)(const char *ifname, struct vendor_ie *ies, int *num_ies)</b>\n
 *	@brief             Get list of vendor information elements
 *	@param[in] ifname  interface name
 *	@param[out] ies    array of struct vendor_ie
 *	@param[out] num    array size (number of elements)
 *
 * <b>int (*get_param)(const char *ifname, const char *param, int *len, void *val)</b>\n
 *	@brief             Get AP parameter value(s).
 *	@param[in] ifname  interface name
 *	@param[in] param   parameter name
 *	@param[out] len    length of the returned value
 *	@param[out] val    parameter value
 *
 * <b>int (*set_param)(const char *ifname, const char *param, int len, void *val)</b>\n
 *	@brief             Set AP parameter value(s).
 *	@param[in] ifname  interface name
 *	@param[in] param   parameter name
 *	@param[in] len     length of the parameter
 *	@param[in] val     value of parameter
 *
 * <b>int (*vendor_cmd)(const char *ifname, uint32_t vid, uint32_t subcmd,
 *			uint8_t *in, int inlen, uint8_t *out, int *outlen)</b>\n
 *	@brief             Vendor specific command
 *	@param[in] ifname  interface name
 *	@param[in] vid     vendor id (OUI)
 *	@param[in] subcmd  (sub)command
 *	@param[in] in      input parameter
 *	@param[in] inlen   length of the input parameter
 *	@param[out] out    output parameter
 *	@param[out] outlen length of the output parameter
 *
 * <b>int (*subscribe_frame)(const char *ifname, uint8_t type, uint8_t stype)</b>\n
 *	@brief             Subscribe for received frames
 *	@param[in] name    interface name
 *	@param[in] type    frame type as in IEEE802.11 Std.
 *	@param[in] stype   frame sub-type as in IEEE802.11 Std.
 *
 * <b>int (*unsubscribe_frame)(const char *ifname, uint8_t type, uint8_t stype)</b>\n
 *	@brief             Unsubscribe for received frames
 *	@param[in] name    interface name
 *	@param[in] type    frame type as in IEEE802.11 Std.
 *	@param[in] stype   frame sub-type as in IEEE802.11 Std.
 *
 * <b>int (*set_4addr)(const char *ifname, bool enable)</b>\n
 *	@brief             Enable or disable 4-address WDS mode.
 *	@param[in] ifname  interface name
 *	@param[in] enable  enable = 1, else disable.
 *
 * <b>int (*set_vlan)(const char *ifname, struct vlan_param vlan)</b>\n
 *	@brief             Set VLAN link.
 *	@param[in] ifname  interface name
 *	@param[in] vlan    vlan parameters
 *
 * <b>int (*ap_info)(const char *ifname, struct wifi_ap *ap)</b>\n
 *	@brief             Get detailed AP information
 *	@param[in] ifname  interface name
 *	@param[out] ap     ap information
 *
 * <b>int (*get_bssid)(const char *ifname, uint8_t *bssid)</b>\n
 *	@brief             Get BSSID
 *	@param[in] ifname  interface name
 *	@param[out] bssid  BSSID buffer (6 bytes)
 *
 * <b>int (*get_ssid)(const char *ifname, char *ssid)</b>\n
 *	@brief             Get SSID
 *	@param[in] ifname  interface name
 *	@param[out] ssid   SSID buffer
 *
 * <b>int (*get_stats)(const char *ifname, struct wifi_ap_stats *s)</b>\n
 *	@brief             Get statistics
 *	@param[in] ifname  interface name
 *	@param[out] s      wifi_ap_stats structure
 *
 * <b>int (*get_beacon_ies)(const char *ifname, uint8_t *ies, int *len)</b>\n
 *	@brief             Get Beacon frame information elements
 *	@param[in] ifname  interface name
 *	@param[out] ies    information elements buffer
 *	@param[out] len    length of information elements buffer
 *
 * <b>int (*get_assoclist)(const char *ifname, uint8_t *stas, int *num_stas)</b>\n
 *	@brief             Get STA association list
 *	@param[in] ifname  interface name
 *	@param[out] stas   macaddresses of STAs
 *	@param[out] num_stas  number of STAs
 *
 * <b>int (*get_sta_info)(const char *ifname, uint8_t *addr, struct wifi_sta *info)</b>\n
 *	@brief             Get STA information
 *	@param[in] ifname  interface name
 *	@param[in] addr    macaddress of STA
 *	@param[out] info   STA information
 *
 * <b>int (*get_sta_stats)(const char *ifname, uint8_t *addr, struct wifi_sta_stats *s)</b>\n
 *	@brief             Get STA statistics
 *	@param[in] ifname  interface name
 *	@param[in] addr    macaddress of STA
 *	@param[out] s      STA counters
 *
 * <b>int (*disconnect_sta)(const char *ifname, uint8_t *sta)</b>\n
 *	@brief             Disconnect STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of STA
 *	@param[in] reason  disconnect reason code as in IEEE802.11 Std
 *
 * <b>int (*restrict_sta)(const char *ifname, uint8_t *sta, int enable)</b>\n
 *	@brief             Assoc-control STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of STA
 *	@param[in] enable  enable (= 1) or disable (= 0) assoc-control
 *
 * <b>int (*monitor_sta)(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg)</b>\n
 *	@brief             Monitor STA frames
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of STA
 *	@param[in] cfg     monitor STA config
 *
 * <b>int (*get_monitor_sta)(const char *ifname, uint8_t *sta, struct wifi_monsta *sta)</b>\n
 *	@brief             Get monitored STA information
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of STA
 *	@param[out] mon    wifi_monsta structure
 *
 * <b>int (*get_monitor_stas)(const char *ifname, struct wifi_monsta *stas, int *num)</b>\n
 *	@brief             Get monitored STA information
 *	@param[in] ifname  interface name
 *	@param[out] stas   array of struct wifi_monsta
 *	@param[out] num    array size (number of wifi_monsta elements)
 *
 * <b>int (*add_neighbor)(const char *ifname, struct nbr nbr)</b>\n
 *	@brief             Add a 802.11k neighbor entry
 *	@param[in] ifname  interface name
 *	@param[in] nbr     nbr structure
 *
 * <b>int (*del_neighbor)(const char *ifname, unsigned char *bssid)</b>\n
 *	@brief             Delete a 802.11k neighbor entry
 *	@param[in] ifname  interface name
 *	@param[in] bssid   Bssid of the neighbor
 *
 * <b>int (*get_neighbor_list)(const char *ifname, struct nbr *nbr, int *nr)</b>\n
 *	@brief             Get 802.11k neighbor list
 *	@param[in] ifname  interface name
 *	@param[out] nbr    array of struct nbr
 *	@param[out] nr     number of array entries
 *
 * <b>int (*req_beacon_report)(const char *ifname, uint8_t *sta)</b>\n
 *	@brief             Request 802.11k Beacon Report from a STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of the STA
 *
 * <b>int (*get_beacon_report)(const char *ifname, uint8_t *sta,
 *					struct sta_nbr *snbr, int *nr)</b>\n
 *	@brief             Get 802.11k Beacon Report received from a STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of the STA
 *	@param[out] snbr   array of sta_nbr structures
 *	@param[out] nr     number of array entries
 *
 * <b>int (*req_bss_transition)(const char *ifname, unsigned char *sta,
 *		int bsss_nr, unsigned char *bsss, unsigned int tmo)</b>\n
 *	@brief             [Deprecated] Request 802.11v BSS transition to a STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of the STA
 *	@param[in] bsss_nr number of neighbor bssids
 *	@param[in] bsss    array of neighbor bssids
 *	@param[in] tmo     timeout (secs) until this request is valid
 *
 * <b>int (*req_btm)(const char *ifname, unsigned char *sta,
 *		int bsss_nr, unsigned char *bsss, struct wifi_btmreq *b)</b>\n
 *	@brief             Request 802.11v BSS transition to a STA
 *	@param[in] ifname  interface name
 *	@param[in] sta     macaddress of the STA
 *	@param[in] bsss_nr number of neighbor bssids
 *	@param[in] bsss    array of neighbor bssids
 *	@param[in] b       additional request parameters
 *
 * <b>int (*get_11rkeys)(const char *ifname, unsigned char *sta, uint8_t *r1khid)</b>\n
 *	@brief             Get 802.11r keys
 *	@param[in] ifname  interface name
 *	@param[in] sta:    macaddress of the STA
 *	@param[out] rikhid:  R1KHID
 *
 * <b>int (*set_11rkeys)(const char *ifname, struct fbt_keys *fk)</b>\n
 *	@brief             Set 802.11r keys
 *	@param[in] ifname  interface name
 *	@param[in] fk      fbt_keys struct
 *
 * <b>int (*sta_info)(const char *ifname, struct wifi_sta *sta)</b>\n
 *	@brief             Get detailed STA information
 *	@param[in] ifname  interface name
 *	@param[out] sta    STA information
 *
 * <b>int (*sta_get_stats)(const char *ifname, struct wifi_sta_stats *s)</b>\n
 *	@brief             Get STA interface statistics
 *	@param[in] ifname  interface name
 *	@param[out] s      STA interface statistics
 *
 * <b>int (*sta_get_ap_info)(const char *ifname, struct wifi_bss *info)</b>\n
 *	@brief             Get BSS information of the STA's AP
 *	@param[in] ifname  interface name
 *	@param[out] info   BSS information of STA's AP
 *
 * <b>int (*sta_disconnect_ap)(const char *ifname, uint32_t reason)</b>\n
 *	@brief             Disconnect from STA's AP
 *	@param[in] ifname  interface name
 *	@param[in] reason  disconnection reason code as in IEEE802.11 Std.
 *
 */
struct wifi_iface_ops {
	/*
	 * common ops
	 */
	int (*start_wps)(const char *ifname, struct wps_param wps);
	int (*stop_wps)(const char *ifname);
	int (*get_wps_status)(const char *ifname, enum wps_status *s);
	int (*get_wps_pin)(const char *ifname, unsigned long *pin);
	int (*set_wps_pin)(const char *ifname, unsigned long pin);
	int (*get_wps_device_info)(const char *ifname, struct wps_device *info);

	int (*get_caps)(const char *ifname, struct wifi_caps *caps);
	int (*get_mode)(const char *ifname, enum wifi_mode *mode);
	int (*get_security)(const char *ifname, uint32_t *auth, uint32_t *enc);

	int (*add_vendor_ie)(const char *ifname, struct vendor_iereq *req);
	int (*del_vendor_ie)(const char *ifname, struct vendor_iereq *req);
	int (*get_vendor_ies)(const char *ifname, struct vendor_ie *ies, int *num_ies);

	int (*get_param)(const char *ifname, const char *param, int *len, void *val);
	int (*set_param)(const char *ifname, const char *param, int len, void *val);
	int (*vendor_cmd)(const char *ifname, uint32_t vid, uint32_t subcmd,
			uint8_t *in, int inlen, uint8_t *out, int *outlen);

	int (*subscribe_frame)(const char *ifname, uint8_t type, uint8_t stype);
	int (*unsubscribe_frame)(const char *ifname, uint8_t type, uint8_t stype);
	int (*set_4addr)(const char *ifname, bool enable);
	int (*set_vlan)(const char *ifname, struct vlan_param vlan);


	/*
	 * ap interface ops
	 */
	int (*ap_info)(const char *name, struct wifi_ap *ap);
	int (*get_bssid)(const char *ifname, uint8_t *bssid);
	int (*get_ssid)(const char *ifname, char *ssid);
	int (*get_stats)(const char *ifname, struct wifi_ap_stats *s);
	int (*get_beacon_ies)(const char *ifname, uint8_t *ies, int *len);

	int (*get_assoclist)(const char *ifname, uint8_t *stas, int *num_stas);
	int (*get_sta_info)(const char *ifname, uint8_t *addr, struct wifi_sta *info);
	int (*get_sta_stats)(const char *ifname, uint8_t *addr, struct wifi_sta_stats *s);
	int (*disconnect_sta)(const char *ifname, uint8_t *sta, uint16_t reason);
	int (*restrict_sta)(const char *ifname, uint8_t *sta, int enable);
	int (*monitor_sta)(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg);
	int (*get_monitor_sta)(const char *ifname, uint8_t *sta, struct wifi_monsta *mon);
	int (*get_monitor_stas)(const char *ifname, struct wifi_monsta *stas, int *num);

	int (*add_neighbor)(const char *ifname, struct nbr nbr);
	int (*del_neighbor)(const char *ifname, unsigned char *bssid);
	int (*get_neighbor_list)(const char *ifname, struct nbr *nbr, int *nr);
	int (*req_beacon_report)(const char *ifname, uint8_t *sta);
	int (*get_beacon_report)(const char *ifname, uint8_t *sta,
					struct sta_nbr *snbr, int *nr);

	int (*req_bss_transition)(const char *ifname, unsigned char *sta,
			int bsss_nr, unsigned char *bsss, unsigned int tmo);

	int (*req_btm)(const char *ifname, unsigned char *sta,
			int bsss_nr, uint8_t *bsss, struct wifi_btmreq *b);

	int (*get_11rkeys)(const char *ifname, unsigned char *sta, uint8_t *r1khid);
	int (*set_11rkeys)(const char *ifname, struct fbt_keys *fk);

	int (*chan_switch)(const char *ifname, struct chan_switch_param *param);

	/*
	 * sta interface ops
	 */
	int (*sta_info)(const char *name, struct wifi_sta *sta);
	int (*sta_get_stats)(const char *ifname, struct wifi_sta_stats *s);
	int (*sta_get_ap_info)(const char *ifname, struct wifi_bss *info);
	int (*sta_disconnect_ap)(const char *ifname, uint32_t reason);
};

/** struct wifi_metainfo - meta information about wifi module */
struct wifi_metainfo {
	char vendor_id[8];		/**< 0xvvvv */
	char device_id[8];		/**< 0xdddd */
	char drv_data[128];		/**< driver name + version info */
	char fw_data[128];		/**< firmware name + version */
};


struct wifi_driver {
	const char *name;
	const char **(*get_apis)(const char *name);
	int (*info)(const char *name, struct wifi_metainfo *info);
	int (*radio_list)(struct radio_entry *radio, int *num);
	struct wifi_radio_ops radio;
	struct wifi_iface_ops iface;
	int (*register_event)(const char *ifname, struct event_struct *ev,
					void **evhandle);
	int (*unregister_event)(const char *ifname, void *evhandle);
	int (*recv_event)(const char *ifname, void *evhandle);
	const char *(*get_version)(void);
};

#define RADIO_OP(_p)	radio._p
#define IFACE_OP(_p)	iface._p

#define scan			RADIO_OP(scan)
#define get_scan_results	RADIO_OP(get_scan_results)
#define get_bss_scan_result	RADIO_OP(get_bss_scan_result)
#define get_channel		RADIO_OP(get_channel)
#define set_channel		RADIO_OP(set_channel)
#define acs			RADIO_OP(acs)
#define get_bandwidth		RADIO_OP(get_bandwidth)
#define get_noise		RADIO_OP(get_noise)
#define get_maxrate		RADIO_OP(get_maxrate)
#define get_country		RADIO_OP(get_country)
#define get_supp_channels	RADIO_OP(get_supp_channels)
#define get_oper_channels	RADIO_OP(get_oper_channels)
#define get_oper_stds		RADIO_OP(get_oper_stds)
#define get_supp_band		RADIO_OP(get_supp_band)
#define get_oper_band		RADIO_OP(get_oper_band)
#define get_supp_opclass	RADIO_OP(get_supp_opclass)
#define get_curr_opclass	RADIO_OP(get_curr_opclass)
#define add_iface		RADIO_OP(add_iface)
#define del_iface		RADIO_OP(del_iface)
#define list_iface		RADIO_OP(list_iface)

#define get_bssid		IFACE_OP(get_bssid)
#define get_ssid		IFACE_OP(get_ssid)
#define disconnect_sta		IFACE_OP(disconnect_sta)
#define restrict_sta		IFACE_OP(restrict_sta)
#define monitor_sta		IFACE_OP(monitor_sta)
#define get_monitor_sta		IFACE_OP(get_monitor_sta)
#define get_monitor_stas	IFACE_OP(get_monitor_stas)
#define vendor_cmd		IFACE_OP(vendor_cmd)
#define get_assoclist		IFACE_OP(get_assoclist)
#define get_sta_info		IFACE_OP(get_sta_info)
#define get_sta_stats		IFACE_OP(get_sta_stats)
#define set_4addr		IFACE_OP(set_4addr)
#define set_vlan		IFACE_OP(set_vlan)
#define get_beacon_ies		IFACE_OP(get_beacon_ies)
#define chan_switch		IFACE_OP(chan_switch)


/* List of the APIs this library provides */

int wifi_driver_info(const char *name, struct wifi_metainfo *info);

/* WiFi radio APIs */
int wifi_radio_list(struct radio_entry *radio, int *num);
int wifi_radio_info(const char *name, struct wifi_radio *radio);
int wifi_get_supp_band(const char *name, uint32_t *bands);
int wifi_get_oper_band(const char *name, enum wifi_band *band);
int wifi_radio_get_caps(const char *name, struct wifi_caps *caps);
int wifi_get_supp_stds(const char *name, uint8_t *std);
int wifi_get_oper_stds(const char *name, uint8_t *std);

int wifi_get_country(const char *name, char *alpha2);
int wifi_get_channel(const char *ifname, uint32_t *channel, enum wifi_bw *bw);
int wifi_set_channel(const char *ifname, uint32_t channel, enum wifi_bw bw);
int wifi_get_supp_channels(const char *name, uint32_t *chlist, int *num,
      	const char *alpha2, enum wifi_band f, enum wifi_bw b);
int wifi_get_oper_channels(const char *name, uint32_t *chlist, int *num,
      	const char *alpha2, enum wifi_band f, enum wifi_bw b);

int wifi_get_supp_opclass(const char *name, int *num_opclass,
						struct wifi_opclass *o);

int wifi_get_curr_opclass(const char *name, struct wifi_opclass *o);

int wifi_get_bandwidth(const char *name, enum wifi_bw *bw);
int wifi_get_maxrate(const char *name, unsigned long *rate);
int wifi_radio_get_basic_rates(const char *name, int *num, uint32_t *rates);
int wifi_radio_get_oper_rates(const char *name, int *num, uint32_t *rates);
int wifi_radio_get_supp_rates(const char *name, int *num, uint32_t *rates);
int wifi_radio_get_stats(const char *ifname, struct wifi_radio_stats *s);

int wifi_scan(const char *name, struct scan_param *p);
int wifi_get_scan_results(const char *name, struct wifi_bss *bsss, int *num);
int wifi_get_bss_scan_result(const char *name, uint8_t *bssid,
						struct wifi_bss_detail *b);

int wifi_get_noise(const char *ifname, int *noise);

int wifi_acs(const char *name, struct acs_param *p);

int wifi_radio_get_param(const char *name, const char *param, int *len, void *val);
int wifi_radio_set_param(const char *name, const char *param, int len, void *val);

int wifi_add_iface(const char *name, enum wifi_mode m, char *argv[]);
int wifi_del_iface(const char *name, const char *ifname);
int wifi_list_iface(const char *name, struct iface_entry *iface, int *num);


/** WiFi interface APIs */
int wifi_start_wps(const char *ifname, struct wps_param wps);
int wifi_stop_wps(const char *ifname);
int wifi_get_wps_status(const char *ifname, enum wps_status *s);
int wifi_get_wps_pin(const char *ifname, unsigned long *pin);
int wifi_set_wps_pin(const char *ifname, unsigned long pin);
int wifi_get_wps_device_info(const char *ifname, struct wps_device *info);

int wifi_get_caps(const char *ifname, struct wifi_caps *caps);
int wifi_get_mode(const char *ifname, enum wifi_mode *mode);
int wifi_get_security(const char *ifname, uint32_t *auth, uint32_t *enc);

int wifi_add_vendor_ie(const char *ifname, struct vendor_iereq *req);
int wifi_del_vendor_ie(const char *ifname, struct vendor_iereq *req);
int wifi_get_vendor_ies(const char *ifname, struct vendor_ie *ies, int *num_ies);

int wifi_get_param(const char *name, const char *param, int *len, void *val);
int wifi_set_param(const char *name, const char *param, int len, void *val);
int wifi_vendor_cmd(const char *ifname, uint32_t vid, uint32_t subcmd,
			uint8_t *in, int ilen, uint8_t *out, int *olen);

int wifi_subscribe_frame(const char *ifname, uint8_t type, uint8_t stype);
int wifi_unsubscribe_frame(const char *ifname, uint8_t type, uint8_t stype);
int wifi_set_4addr(const char *ifname, bool enable);
int wifi_set_vlan(const char *ifname, struct vlan_param vlan);


/* WiFi AP interface specific APIs */
int wifi_ap_info(const char *name, struct wifi_ap *ap);
int wifi_get_bssid(const char *ifname, uint8_t *bssid);
int wifi_get_ssid(const char *ifname, char *ssid);
int wifi_ap_get_stats(const char *ifname, struct wifi_ap_stats *s);
int wifi_get_beacon_ies(const char *ifname, uint8_t *ies, int *len);

int wifi_get_assoclist(const char *ifname, uint8_t *stas, int *num_stas);
int wifi_get_sta_info(const char *ifname, uint8_t *addr, struct wifi_sta *info);
int wifi_get_sta_stats(const char *ifname, uint8_t *addr, struct wifi_sta_stats *s);
int wifi_disconnect_sta(const char *ifname, uint8_t *sta, uint16_t reason);
int wifi_restrict_sta(const char *ifname, uint8_t *sta, int enable);
int wifi_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta_config *cfg);
int wifi_get_monitor_sta(const char *ifname, uint8_t *sta, struct wifi_monsta *mon);
int wifi_get_monitor_stas(const char *ifname, struct wifi_monsta *stas, int *num);

int wifi_add_neighbor(const char *ifname, struct nbr nbr);
int wifi_del_neighbor(const char *ifname, unsigned char *bssid);
int wifi_get_neighbor_list(const char *ifname, struct nbr *nbrs, int *nr);
int wifi_req_beacon_report(const char *ifname, uint8_t *sta);
int wifi_get_beacon_report(const char *ifname, uint8_t *sta,
				struct sta_nbr *snbrs, int *nr);
int wifi_req_bss_transition(const char *ifname, unsigned char *sta,
			int bsss_nr, unsigned char *bsss,
			unsigned int tmo);

int wifi_req_btm(const char *ifname, uint8_t *sta, int bsss_nr, uint8_t *bsss,
					struct wifi_btmreq *b);

int wifi_get_11rkeys(const char *ifname, unsigned char *sta, uint8_t *r1khid);
int wifi_set_11rkeys(const char *ifname, struct fbt_keys *fk);
int wifi_chan_switch(const char *ifname, struct chan_switch_param *param);


/* WiFi STA interface APIs */
int wifi_sta_info(const char *ifname, struct wifi_sta *sta);
int wifi_sta_get_stats(const char *ifname, struct wifi_sta_stats *s);
int wifi_sta_get_ap_info(const char *ifname, struct wifi_bss *info);
int wifi_sta_disconnect_ap(const char *ifname, uint32_t reason);


/* WiFi events */
enum wifi_event_type {
	WIFI_EVENT_UNDEFINED,
	WIFI_EVENT_VENDOR = 1,
	WIFI_EVENT_SCAN_START,
	WIFI_EVENT_SCAN_END,
	WIFI_EVENT_STA_ASSOC,
	WIFI_EVENT_STA_DISASSOC,
	WIFI_EVENT_CAC_START,
	WIFI_EVENT_CAC_END,
	WIFI_EVENT_RADAR_DETECT,
	WIFI_EVENT_CHANNEL,
	WIFI_EVENT_WPS_OVERLAP,
	WIFI_EVENT_WPS_SUCCESS,
	WIFI_EVENT_BTM_RESPONSE,
	WIFI_EVENT_FRAME,

	/* add new types here */

	WIFI_EVENT_DEFAULT = 99,   /* last resort */
};


int wifi_register_event(const char *ifname, struct event_struct *ev,
							void **evhandle);
int wifi_unregister_event(const char *ifname, void *evhandle);
int wifi_recv_event(const char *ifname, void *evhandle);


/** vendor agnostic wifi APIs */

/** Get interface state */
int wifi_get_ifstatus(const char *ifname, ifstatus_t *f);

/** Get interface operational state */
int wifi_get_ifoperstatus(const char *ifname, ifopstatus_t *opstatus);

/** Get valid channels for given country, band and bandwidth */
int wifi_get_valid_channels(const char *ifname, enum wifi_band b,
					enum wifi_bw bw, const char *cc,
					uint32_t *chlist, int *n);


/** Get current operating class of the wifi radio */
int wifi_get_opclass(const char *name, struct wifi_opclass *o);

/** Get operating class for current channel and HT20 */
int wifi_get_opclass_ht20(const char *name, struct wifi_opclass *o);

/** Get list of operating classes supported by the wifi radio */
int wifi_get_supported_opclass(const char *name, int *num_opclass,
						struct wifi_opclass *o);

/** Generate a random wps pin */
int wifi_generate_wps_pin(unsigned long *pin);

/** Check if a wps pin is valid. Return 1 if valid; else 0 */
int wifi_is_wps_pin_valid(unsigned long pin);

/** Helper functions to convert frequency to/from channel number */
int wifi_channel_to_freq(int chan);
int wifi_freq_to_channel(int freq);

/** Iterator for information elements */
#define wifi_foreach_ie(e, _iebuf, _len)				\
	for ((e) = (_iebuf);						\
		(e)[1] && (_iebuf) + (_len) - (e) - 2 - (e)[1] >= 0;	\
		(e) += 2 + (e)[1])

/* Get list of APIs that this library supports */
const char **libwifi_get_apis(char *name);

/* Get library version */
const char *libwifi_get_version(void);

/* Helper function to check if an API is defined by a driver implementation */
extern int libwifi_supports(const char *ifname, char *api_name);


#ifdef __cplusplus
}
#endif

#endif /* WIFI_H */
